@@ -31,6 +31,9 @@ void meta_core_get_client_size (Display *xdisplay,
                                 int     *width,
                                 int     *height);
 
+Window meta_core_get_client_xwindow (Display *xdisplay,
+                                     Window   frame_xwindow);
+
 MetaFrameFlags meta_core_get_frame_flags (Display *xdisplay,
                                           Window   frame_xwindow);
 MetaFrameType  meta_core_get_frame_type   (Display *xdisplay,
@@ -102,9 +105,9 @@ int meta_core_get_num_workspaces (Screen  *xscreen);
 int meta_core_get_active_workspace (Screen *xscreen);
 int meta_core_get_frame_workspace (Display *xdisplay,
                                    Window frame_xwindow);
-char* meta_core_get_workspace_name_with_index (Display *xdisplay,
-                                               Window xroot,
-                                               int    index);
+const char* meta_core_get_workspace_name_with_index (Display *xdisplay,
+                                                     Window xroot,
+                                                     int    index);
 
 void  meta_core_get_frame_extents   (Display        *xdisplay,
                                      Window          frame_xwindow,
